from collections import defaultdict
from typing import Any, Dict, List

import numpy as np

import fast_bfmatcher.matchers as matchers
from fast_bfmatcher.extra.cv import OpenCVL2CCBFMatcher, OpenCVL2RTBFMatcher
from fast_bfmatcher.matching_ops import find_cross_check_matches
from fast_bfmatcher.utils import measuretime


def benchmark(name: str, method, steps: int = 200, warmup: int = 10):

    with measuretime(f"{name:20} warmup", num_steps=steps, log=False):
        for _ in range(warmup):
            method()

    with measuretime(f"{name:20} calls ", num_steps=steps):
        for _ in range(steps):
            method()


def benchmark_cc_matchers(
    steps: int = 100, warmup: int = 5, num_kpts: int = 2000, dim: int = 128
):

    benchmark_tf = True
    try:
        from fast_bfmatcher.extra.tf import TFL2CCBFMatcher

        tf_matcher = TFL2CCBFMatcher()
    except Exception as e:
        print(f"Skipping tensorflow benchmark, got error: {e}")
        benchmark_tf = False

    X = np.random.randint(0, 255, size=(num_kpts, dim)).astype(np.float32)
    Y = np.random.randint(0, 255, size=(num_kpts, dim)).astype(np.float32)
    D = np.random.randint(0, 255, size=(num_kpts, num_kpts)).astype(np.float32)

    fast_matcher = matchers.FastL2CCBFMatcher()

    cv_matcher = OpenCVL2CCBFMatcher()

    from fast_bfmatcher.extra.np import NumpyL2CCBFMatcher

    np_matcher = NumpyL2CCBFMatcher()

    print("\n>> Benchmarking matchers ...")
    benchmark("fast", lambda: fast_matcher.match(X, Y), steps=steps, warmup=warmup)
    benchmark("opencv", lambda: cv_matcher.match(X, Y), steps=steps, warmup=warmup)
    benchmark("numpy", lambda: np_matcher.match(X, Y), steps=steps, warmup=warmup)

    if benchmark_tf:
        benchmark(
            "tensorflow", lambda: tf_matcher.match(X, Y), steps=steps, warmup=warmup
        )

    print("\n>> Benchmarking distance matrix computation ...")

    benchmark(
        "fast", lambda: matchers.l2_distance_matrix(X, Y), steps=steps, warmup=warmup
    )
    benchmark(
        "numpy",
        lambda: NumpyL2CCBFMatcher.l2_distance_matrix(X, Y),
        steps=steps,
        warmup=warmup,
    )

    if benchmark_tf:
        benchmark(
            "tensorflow",
            lambda: TFL2CCBFMatcher.l2_distance_matrix(X, Y),
            steps=steps,
            warmup=warmup,
        )

    print("\n>> Benchmarking find row col min values and indices ...")

    def _find_row_col_min_values_np(x):
        row_indices_np = x.argmin(1)
        row_values_np = x.min(1)
        col_values_np = x.min(0)
        return row_indices_np, row_values_np, col_values_np

    benchmark("fast", lambda: find_cross_check_matches(D), steps=steps, warmup=warmup)
    benchmark(
        "numpy", lambda: _find_row_col_min_values_np(D), steps=steps, warmup=warmup
    )

    if benchmark_tf:
        benchmark(
            "tensorflow",
            lambda: TFL2CCBFMatcher.find_row_col_min_values(D),
            steps=steps,
            warmup=warmup,
        )


def benchmark_cc_rt_size_scan(
    max_size: int = 10000, random_sampler_per_size: int = 20
) -> Dict[str, List[Any]]:
    """
    Measures matching time between two random int vectors of size [size, 128] in
    function of size parameter.

    NOTE: Before run set number of threads
                os.environ["BLIS_NUM_THREADS"] = "8"

    Args:
        max_size: maximum size of the tested vector
        random_sampler_per_size: number of random sampler per each size value

    Returns:
        metrics: dict with benchmark metrics
    """

    def _test_match(matcher: matchers.Matcher, name: str) -> Dict[str, List[Any]]:

        metrics = defaultdict(list)
        for n in range(100, max_size, 500):
            X = np.random.randint(0, 255, size=(n, 128))
            Y = np.random.randint(0, 255, size=(n, 128))
            # warmup
            for _ in range(5):
                matcher.match(X, Y)

            with measuretime(f"{name} N={n}") as dt:
                for _ in range(random_sampler_per_size):
                    matcher.match(X, Y)

            metrics[name].append(1000 * dt.seconds / random_sampler_per_size)
            metrics["size"].append(n)

        return metrics

    fast_matcher_rt = matchers.FastL2RTBFMatcher(ratio=0.7)
    fast_matcher_cc = matchers.FastL2CCBFMatcher()

    cv_matcher_rt = OpenCVL2RTBFMatcher(0.7)
    cv_matcher_cc = OpenCVL2CCBFMatcher()

    fs_rt_metrics = _test_match(fast_matcher_rt, "fast-rt")
    fs_cc_metrics = _test_match(fast_matcher_cc, "fast-cc")
    cv_rt_metrics = _test_match(cv_matcher_rt, "opencv-rt")
    cv_cc_metrics = _test_match(cv_matcher_cc, "opencv-cc")

    return {**fs_rt_metrics, **fs_cc_metrics, **cv_rt_metrics, **cv_cc_metrics}



import cv2
import numpy as np

from fast_bfmatcher.matchers import Matcher, MatchResult


class OpenCVL2CCBFMatcher(Matcher):
    """Reference openCV implementation"""

    def __init__(self):
        self.cv_matcher = cv2.BFMatcher(cv2.NORM_L2, crossCheck=True)

    def match(self, X: np.ndarray, Y: np.ndarray) -> MatchResult:
        X, Y = self.cast_inputs(X, Y)
        matches = self.cv_matcher.knnMatch(X, Y, k=1)
        matches_pairs = []
        distances = []
        for mm in matches:
            if len(mm) == 1:
                match = mm[0]
                row, col = match.queryIdx, match.trainIdx
                matches_pairs.append([row, col])
                distances.append(match.distance)

        return MatchResult(np.array(matches_pairs), np.array(distances))


class OpenCVL2RTBFMatcher(Matcher):
    def __init__(self, ratio: float = 0.7):
        self.ratio = ratio
        self.cv_matcher = cv2.BFMatcher(cv2.NORM_L2, crossCheck=False)

    def match(self, X: np.ndarray, Y: np.ndarray) -> MatchResult:
        X, Y = self.cast_inputs(X, Y)
        matches = self.cv_matcher.knnMatch(X, Y, k=2)
        matches_pairs = []
        distances = []

        for i in range(len(matches)):
            match = matches[i][0]
            if match.distance <= self.ratio * matches[i][1].distance:
                row, col = match.queryIdx, match.trainIdx
                matches_pairs.append([row, col])
                distances.append(match.distance)

        return MatchResult(np.array(matches_pairs), np.array(distances))



import numpy as np
cimport numpy as np

cimport cython
from libc.stdlib cimport malloc, free

np.import_array()


ctypedef np.float32_t Float32_t
ctypedef np.float64_t Float64_t
ctypedef np.uint8_t Uint8_t
ctypedef np.int32_t Int32_t


cdef extern from "fast_ops.h":
    void sum_square_cols(float* X, float *y, int num_rows, int num_cols)
    void sum_row_and_col_vectors(float * row, float *col, float* X, int num_rows, int num_cols)
    void _fast_cross_check_match "fast_cross_check_match"(int *irow, float *vrow, float *vcol, float * X, int num_rows, int num_cols)
    void _fast_ratio_test_match "fast_ratio_test_match"(int *irow, float *vrow, float * X, int num_rows, int num_cols, float ratio)


cdef extern from "blis.h":
    ctypedef enum trans_t:
        BLIS_TRANSPOSE
        BLIS_NO_TRANSPOSE

    void _bli_sgemm "bli_sgemm"(trans_t trans_a, trans_t trans_b,
                        int m, int n, int k, float* alpha,
                        float *a, int rsa, int csa,
                        float *b, int rsb, int csb,
                        float *beta, float *c, int rsc, int csc) nogil


cpdef void blis_sgemm_transpose(
        float alpha,
        float[:, ::1] A,
        float[:, ::1] B,
        float beta,
        float[:, ::1] C
):

    """
    Computes C = α A B^T + β C
    """

    cdef float* A_ptr = &A[0, 0]
    cdef float* B_ptr = &B[0, 0]
    cdef float* C_ptr = &C[0, 0]

    _bli_sgemm(BLIS_NO_TRANSPOSE,BLIS_TRANSPOSE,
            C.shape[0], C.shape[1], A.shape[1],
            &alpha,
            A_ptr, A.shape[1], 1,
            B_ptr, B.shape[1], 1,
            &beta,
            C_ptr, C.shape[1], 1)


cpdef void l2_distance_matrix(float[:, ::1] A, float[:, ::1] B, float[:, ::1] C):
    """
    Computes squared euclidean distance matrix between arrays A and B
        
    Args:
        A: array of shape [N, C] 
        B: array of shape [M, C]
        C: output array of shape [N, M]

    """
    cdef float* A_ptr = &A[0, 0]
    cdef float* B_ptr = &B[0, 0]
    cdef float* C_ptr = &C[0, 0]

    cdef int a_num_rows = A.shape[0]
    cdef int a_num_cols = A.shape[1]
    cdef int b_num_rows = B.shape[0]
    cdef int b_num_cols = B.shape[1]

    cdef float *a_sq = <float*> malloc(a_num_rows * sizeof(float))
    cdef float *b_sq = <float*> malloc(b_num_rows * sizeof(float))

    try:
        sum_square_cols(A_ptr, a_sq, a_num_rows, a_num_cols)
        sum_square_cols(B_ptr, b_sq, b_num_rows, b_num_cols)
        sum_row_and_col_vectors(a_sq, b_sq, C_ptr, a_num_rows, b_num_rows)
        blis_sgemm_transpose(-2, A, B, 1, C)

    finally:
        free(a_sq)
        free(b_sq)


cpdef l2_cross_check_matcher(float[:, ::1] A, float[:, ::1] B):

    cdef:
        int num_rows = A.shape[0]
        int num_cols = B.shape[0]

        float[:,::1] C = np.zeros((num_rows, num_cols), dtype = np.float32)
        int[::1] row_indices = np.zeros((num_rows,), dtype = np.int32)
        float[::1] row_values = np.zeros((num_rows,), dtype=np.float32)
        float[::1] col_values = np.zeros((num_cols,), dtype=np.float32)

        float *C_ptr = &C[0, 0]

    l2_distance_matrix(A, B, C)

    _fast_cross_check_match(
        &row_indices[0], &row_values[0], &col_values[0],
        C_ptr, num_rows, num_cols
    )

    row_index = np.arange(0, A.shape[0])
    cross_checked = row_values == np.array(col_values)[row_indices]
    rows = row_index[cross_checked]
    cols = np.array(row_indices)[cross_checked]
    distances = np.array(row_values)[cross_checked]
    indices = np.transpose(np.stack([rows, cols]))
    distances = np.sqrt(np.maximum(0, distances))
    return indices, distances


cpdef find_cross_check_matches(float[:, ::1] X):
    """
    
    Args:
        X: [N, M] float32 distance matrix  

    """

    cdef:
        int num_rows = X.shape[0]
        int num_cols = X.shape[1]

        int[::1] row_indices = np.zeros((num_rows,), dtype = np.int32)
        float[::1] row_values = np.zeros((num_rows,), dtype=np.float32)
        float[::1] col_values = np.zeros((num_cols,), dtype=np.float32)

        float *X_ptr = &X[0, 0]

    _fast_cross_check_match(
        &row_indices[0], &row_values[0], &col_values[0],
        X_ptr, num_rows, num_cols
    )

    return row_indices, row_values, col_values


cpdef l2_ratio_test_matcher(float[:, ::1] A, float[:, ::1] B, float ratio):

    cdef:
        int num_rows = A.shape[0]
        int num_cols = B.shape[0]

        float[:,::1] D = np.zeros((num_rows, num_cols), dtype = np.float32)
        int[::1] row_indices = np.zeros((num_rows,), dtype = np.int32)
        float[::1] row_values = np.zeros((num_rows,), dtype=np.float32)
        float[::1] col_values = np.zeros((num_cols,), dtype=np.float32)

        # pointer to distance matrix
        float *D_ptr = &D[0, 0]
        # use ratio^2 because D is squared distance matrix
        float squared_ratio = ratio ** 2

    l2_distance_matrix(A, B, D)

    _fast_ratio_test_match(
        &row_indices[0], &row_values[0],
        D_ptr, num_rows, num_cols,
        squared_ratio
    )

    row_index = np.arange(0, A.shape[0])
    valid_matches = np.array(row_indices) != -1

    rows = row_index[valid_matches]
    cols = np.array(row_indices)[valid_matches]

    distances = np.array(row_values)[valid_matches]
    indices = np.transpose(np.stack([rows, cols]))
    distances = np.sqrt(np.maximum(0, distances))
    return indices, distances


cpdef find_ratio_test_matches(float[:, ::1] X, float ratio = 0.7):
    """

    Args:
        X: [N, M] float32 distance matrix  
        ratio: ratio test threshold for the second nearest neighbour

    """

    cdef:
        int num_rows = X.shape[0]
        int num_cols = X.shape[1]

        int[::1] row_indices = np.zeros((num_rows,), dtype = np.int32)
        float[::1] row_values = np.zeros((num_rows,), dtype=np.float32)

        float *X_ptr = &X[0, 0]

    _fast_ratio_test_match(
        &row_indices[0], &row_values[0],
        X_ptr, num_rows, num_cols, ratio
    )

    return row_indices, row_values


import numpy as np
import tensorflow as tf

from fast_bfmatcher.matchers import Matcher, MatchResult

if tf.__version__ < "2.0.0":
    if tf.__version__ >= "1.14.0":
        if not tf.compat.v1.executing_eagerly():
            msg = "TFL2BFMatcher requires eager mode to be enabled! Call tf.enable_eager_execution() first."
            print(msg)
            raise ImportError(msg)
    else:
        msg = f"TF Matcher Requires minimum version of Tensorflow is 2.0.0, got: {tf.__version__}"
        print(msg)
        raise ImportError(msg)


@tf.function(input_signature=(tf.TensorSpec(shape=[None, None], dtype=tf.float32),))
def find_row_col_min_values_tf(X: tf.Tensor):
    row_indices = tf.argmin(X, 1)
    row_values = tf.reduce_min(X, 1)
    col_values = tf.reduce_min(X, 0)
    return row_indices, row_values, col_values


@tf.function(
    input_signature=(
        tf.TensorSpec(shape=[None, None], dtype=tf.float32),
        tf.TensorSpec(shape=[None, None], dtype=tf.float32),
    )
)
def l2_distance_matrix_tf(X: tf.Tensor, Y: tf.Tensor) -> tf.Tensor:
    sqnorm1 = tf.reduce_sum(tf.square(X), 1, keepdims=True)
    sqnorm2 = tf.reduce_sum(tf.square(Y), 1, keepdims=True)
    innerprod = tf.matmul(X, Y, transpose_a=False, transpose_b=True)
    return sqnorm1 + tf.transpose(sqnorm2) - 2 * innerprod


class TFL2CCBFMatcher(tf.Module, Matcher):
    def __init__(self, name: str = None):
        super(TFL2CCBFMatcher, self).__init__(name=name)
        self.dtype = tf.float32

    @classmethod
    def cast_input(cls, X: np.ndarray, dtype: np.dtype) -> np.ndarray:
        if X.dtype != dtype:
            X = tf.cast(X, dtype)
        return X

    @classmethod
    def l2_distance_matrix(cls, X: np.ndarray, Y: np.ndarray):
        D = l2_distance_matrix_tf(X, Y)
        return D.numpy()

    @classmethod
    def find_row_col_min_values(cls, X: np.ndarray):
        row_indices, row_values, col_values = find_row_col_min_values_tf(X)
        return row_indices.numpy(), row_values.numpy(), col_values.numpy()

    @tf.function(
        input_signature=(
            tf.TensorSpec(shape=[None, None], dtype=tf.float32),
            tf.TensorSpec(shape=[None, None], dtype=tf.float32),
        )
    )
    def match_fn(self, X: tf.Tensor, Y: tf.Tensor):

        dist_mat = l2_distance_matrix_tf(X, Y)

        row_matches = tf.argmin(dist_mat, 1)
        col_matches = tf.argmin(dist_mat, 0)

        num_rows = tf.cast(tf.shape(row_matches)[0], dtype=row_matches.dtype)

        inverse_row_indices = tf.gather(col_matches, row_matches)
        row_indices = tf.range(0, num_rows, dtype=row_matches.dtype)

        cross_checked = tf.equal(row_indices, inverse_row_indices)
        rows = row_indices[cross_checked]
        cols = row_matches[cross_checked]

        indices = tf.transpose(tf.stack([rows, cols]))
        distances = tf.gather_nd(dist_mat, indices)
        distances = tf.sqrt(tf.maximum(0.0, distances))
        return indices, distances

    def match(self, X: np.ndarray, Y: np.ndarray) -> MatchResult:
        X, Y = self.cast_inputs(X, Y)
        indices, distances = self.match_fn(X, Y)
        return MatchResult(indices.numpy(), distances.numpy())


class TFL2RTBFMatcher(TFL2CCBFMatcher):
    def __init__(self, ratio: float = 0.7, name: str = None):
        super(TFL2CCBFMatcher, self).__init__(name)
        self.ratio = ratio

    @tf.function(
        input_signature=(
            tf.TensorSpec(shape=[None, None], dtype=tf.float32),
            tf.TensorSpec(shape=[None, None], dtype=tf.float32),
        )
    )
    def match_fn(self, X: tf.Tensor, Y: tf.Tensor):

        dist_mat = l2_distance_matrix_tf(X, Y)

        neg_distances, indices = tf.nn.top_k(-dist_mat, 2)
        distances = -neg_distances
        ratio2 = self.ratio ** 2
        is_good = distances[:, 0] < ratio2 * distances[:, 1]

        num_rows = tf.shape(dist_mat)[0]
        row_indices = tf.range(0, num_rows, dtype=indices.dtype)

        rows = row_indices[is_good]
        cols = indices[is_good][:, 0]
        distances = distances[is_good][:, 0]

        indices = tf.transpose(tf.stack([rows, cols]))
        distances = tf.sqrt(tf.maximum(0.0, distances))
        return indices, distances

    def match(self, X: np.ndarray, Y: np.ndarray) -> MatchResult:
        X, Y = self.cast_inputs(X, Y)
        indices, distances = self.match_fn(X, Y)
        return MatchResult(indices.numpy(), distances.numpy())



import numpy as np

from fast_bfmatcher.matchers import Matcher, MatchResult


class NumpyL2CCBFMatcher(Matcher):
    """Reference numpy implementation"""

    @classmethod
    def l2_distance_matrix(cls, X: np.ndarray, Y: np.ndarray):
        sqnorm1 = np.sum(np.square(X), 1, keepdims=True)
        sqnorm2 = np.sum(np.square(Y), 1, keepdims=True)
        innerprod = np.dot(X, Y.T)
        return sqnorm1 + np.transpose(sqnorm2) - 2.0 * innerprod

    def match(self, X: np.ndarray, Y: np.ndarray) -> MatchResult:

        dist_mat = self.l2_distance_matrix(X, Y)

        row_matches = np.argmin(dist_mat, 1)
        col_matches = np.argmin(dist_mat, 0)

        num_rows = row_matches.shape[0]

        inverse_row_indices = col_matches[row_matches]
        row_indices = np.arange(0, num_rows, dtype=row_matches.dtype)

        cross_checked = row_indices == inverse_row_indices
        rows = row_indices[cross_checked]
        cols = row_matches[cross_checked]

        indices = np.transpose(np.stack([rows, cols]))
        distances = dist_mat[rows, cols]
        distances = np.sqrt(np.maximum(0, distances))

        return MatchResult(indices, distances)



import subprocess
from distutils import log
from distutils.errors import DistutilsSetupError
from pathlib import Path
from typing import Union

from setuptools import Extension, dist, setup
from setuptools.command.build_ext import build_ext

dist.Distribution().fetch_build_eggs(["cython", "numpy"])

BLIS_PATH = Path("build/blis")
BLIS_VERSION = "0.9.0"
PACKAGE_NAME = "fast_bfmatcher"
VERSION = "1.3.0"


def run_command(command: str, sources_path: Union[str, Path] = "."):

    sources_path = str(sources_path)
    log.info(f'>> Executing command: "{command}" in directory: {sources_path}\n')

    command = [w.strip() for w in command.split(" ")]
    result = subprocess.run(
        command,
        cwd=sources_path,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=False,
    )

    stdout, stderr = result.stdout.decode(), result.stderr.decode()

    if result.returncode != 0:
        status = stdout or stderr
        raise DistutilsSetupError(
            f"An error occurred while running the command: {command}.\nStatus: {status}"
        )


class compile_blis_and_build_ext(build_ext):
    def build_extension(self, ext: Extension):
        print(f">> Building extension: {ext.name}")

        if not BLIS_PATH.exists():
            Path("build").mkdir(exist_ok=True)
            # clone blis at selected version
            run_command(
                f"git clone --depth 1 --branch {BLIS_VERSION} https://github.com/flame/blis.git",
                "build",
            )
        else:
            log.info(f'>> BLIS_PATH="{BLIS_PATH}" exists, skipping download')

        blis_library_list = list(BLIS_PATH.glob("lib/*/libblis.a"))
        blis_include_list = list(BLIS_PATH.glob("include/*/blis.h"))

        if len(blis_library_list) == 0:
            log.info(">> Building BLIS: https://github.com/flame/blis#getting-started")
            run_command("./configure -t openmp auto", BLIS_PATH)
            run_command("make -j 4", BLIS_PATH)

            blis_library_list = list(BLIS_PATH.glob("lib/*/libblis.a"))
            blis_include_list = list(BLIS_PATH.glob("include/*/blis.h"))

        if len(blis_library_list) == 1:
            blis_library = str(blis_library_list[0])
            blis_include = str(Path(blis_include_list[0]).parent)
        else:
            raise DistutilsSetupError(
                f"Configuration error: expected single BLIS library, got: {blis_library_list}"
            )

        log.info(
            f">> BLIS library has been already built, using library from path: {blis_library}, "
            f"BLIS include path is: {blis_include}"
        )

        ext.include_dirs.append(blis_include)
        ext.extra_link_args.append(blis_library)
        super(compile_blis_and_build_ext, self).build_extension(ext)


def get_include_dir():
    import numpy

    return [numpy.get_include()]


extensions = [
    Extension(
        f"{PACKAGE_NAME}.matching_ops",
        [f"{PACKAGE_NAME}/fast_ops.c", f"{PACKAGE_NAME}/matching_ops.pyx"],
        libraries=["blas"],
        include_dirs=get_include_dir(),
        extra_compile_args=[
            "-Ofast",
            "-march=native",
            "-msse3",
            "-finline-functions",
            "-Wno-unused-function",
            "-Wfatal-errors",
            "-fPIC",
            "-std=c99",
            "-fopenmp",
            "-D_POSIX_C_SOURCE=200112L",
            "-fopt-info-vec-optimized",
        ],
        extra_link_args=[
            "-lm",
            "-fopenmp",
            "-lrt",
        ],
    )
]


def get_extension_modules():
    from Cython.Build import cythonize

    return cythonize(extensions)


setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description="Faster implementation of OpenCV BFMatcher matcher",
    url="https://github.com/kmkolasinski/fast-bfmatcher",
    author="Krzysztof Kolasinski",
    author_email="kmkolasinski@gmail.com",
    license="MIT",
    packages=["fast_bfmatcher", "fast_bfmatcher.extra"],
    include_package_data=True,
    zip_safe=False,
    setup_requires=["cython", "numpy"],
    install_requires=["numpy>=1.16.0"],
    package_data={PACKAGE_NAME: ["matching_ops.pyx", "fast_ops.c", "fast_ops.h"]},
    ext_modules=get_extension_modules(),
    cmdclass={"build_ext": compile_blis_and_build_ext},
)



import dataclasses
from abc import ABC, abstractmethod
from typing import Tuple

import numpy as np

from fast_bfmatcher import matching_ops as mops


@dataclasses.dataclass(frozen=True)
class MatchResult:
    indices: np.ndarray
    """
    Integer array of shape [num_matches, 2] with (i, j) pair
    matches indices between X[i, :] and Y[j, :] descriptors
    """
    distances: np.ndarray
    """
    Float array of shape [num_matches] with computed distances
    between matched (i, j) pairs.
    """

    def __eq__(self, other: "MatchResult"):
        try:
            np.testing.assert_equal(self.indices, other.indices)
            np.testing.assert_almost_equal(self.distances, other.distances, decimal=4)
        except AssertionError:
            return False
        return True


class Matcher(ABC):
    @abstractmethod
    def match(self, X: np.ndarray, Y: np.ndarray) -> MatchResult:
        pass

    @classmethod
    def cast_input(cls, X: np.ndarray, dtype: np.dtype) -> np.ndarray:
        if X.dtype != dtype:
            X = X.astype(dtype)
        return X

    @classmethod
    def cast_inputs(
        cls, X: np.ndarray, Y: np.ndarray, dtype: np.dtype = np.float32
    ) -> Tuple[np.ndarray, np.ndarray]:
        X = cls.cast_input(X, dtype)
        Y = cls.cast_input(Y, dtype)
        return X, Y


class FastL2CCBFMatcher(Matcher):
    """Brute force matcher equivalent to cv2.BFMatcher(NORM_L2, cross_check=True)"""

    def match(self, X: np.ndarray, Y: np.ndarray) -> MatchResult:
        X, Y = self.cast_inputs(X, Y)
        indices, distances = mops.l2_cross_check_matcher(X, Y)
        return MatchResult(indices, distances)


class FastL2RTBFMatcher(Matcher):
    """
    Brute force matcher equivalent to cv2.BFMatcher(NORM_L2, cross_check=False)
    With Lowe's ratio test
    """

    def __init__(self, ratio: float = 0.7):
        self.ratio = ratio

    def match(self, X: np.ndarray, Y: np.ndarray) -> MatchResult:
        X, Y = self.cast_inputs(X, Y)
        indices, distances = mops.l2_ratio_test_matcher(X, Y, ratio=self.ratio)
        return MatchResult(indices, distances)


def l2_distance_matrix(X: np.ndarray, Y: np.ndarray) -> np.ndarray:
    D = np.zeros([X.shape[0], Y.shape[0]], np.float32)
    mops.l2_distance_matrix(X, Y, D)
    return D



import time
from typing import Any, Dict, Optional


class measuretime:
    def __init__(
        self,
        name: str,
        extra: Optional[Dict[str, Any]] = None,
        log: bool = True,
        num_steps: int = 1,
    ):
        self.name = name
        self.extra = extra
        self.log = log
        self.num_steps = num_steps

    @property
    def params(self) -> str:
        if self.extra is None:
            return ""

        params = []
        for k in sorted(self.extra):
            params.append(f"{k}={self.extra[k]}")
        params = ", ".join(params)

        return f"PARAMS: {params}"

    def __enter__(self):
        self.t = time.perf_counter()
        return self

    def __exit__(self, *args, **kwargs):
        self.seconds = time.perf_counter() - self.t
        if self.log:
            print(
                f"{self.name}: {self.num_steps} steps took {self.seconds:6.3f} [s], "
                f"per step {1000 * self.seconds / self.num_steps:6.3f} [ms]"
                f" {self.params}"
            )
